__version__ = "1.0.21"

import kappadata.caching
import kappadata.copying
import kappadata.datasets
import kappadata.loading
import kappadata.transforms
import kappadata.wrappers
import kappadata.wrappers.dataset_wrappers
import kappadata.wrappers.sample_wrappers
from .caching.redis_dataset import RedisDataset
from .caching.shared_dict_dataset import SharedDictDataset
from .collators.base.compose_collator import ComposeCollator
from .collators.base.kd_collator import KDCollator
from .collators.pad_sequences_collator import PadSequencesCollator
from .copying import copy_folder_from_global_to_local
from .datasets.kd_concat_dataset import KDConcatDataset
from .datasets.kd_dataset import KDDataset
from .datasets.kd_subset import KDSubset
from .datasets.kd_wrapper import KDWrapper
from .transforms.base.kd_transform import KDTransform
from .wrappers.dataset_wrappers.class_filter_wrapper import ClassFilterWrapper
from .wrappers.dataset_wrappers.oversampling_wrapper import OversamplingWrapper
from .wrappers.dataset_wrappers.percent_filter_wrapper import PercentFilterWrapper
from .wrappers.dataset_wrappers.repeat_wrapper import RepeatWrapper
from .wrappers.dataset_wrappers.shuffle_wrapper import ShuffleWrapper
from .wrappers.dataset_wrappers.subset_wrapper import SubsetWrapper
from .wrappers.mode_wrapper import ModeWrapper
from .wrappers.sample_wrappers.mixup_wrapper import MixupWrapper



from kappadata.datasets.kd_dataset import KDDataset


class IndexDataset(KDDataset):
    def __init__(self, size):
        super().__init__()
        self.size = size
        self.disposed = False

    @staticmethod
    def getitem_x(idx, _=None):
        return idx

    def __len__(self):
        return self.size

    def dispose(self):
        self.disposed = True



from torch.utils.data import default_collate

from .kd_collator import KDCollator


class ComposeCollator:
    def __init__(self, collators, dataset_mode, return_ctx=False):
        assert isinstance(collators, list) and len(collators) > 0 and all(isinstance(c, KDCollator) for c in collators)
        self.collators = collators
        self.dataset_mode = dataset_mode
        self.return_ctx = return_ctx

        self.default_collate_modes = [c.default_collate_mode for c in collators]
        assert all(dcm in [None, "before", "after"] for dcm in self.default_collate_modes)

    def __call__(self, batch):
        called_default_collate = False
        ctx = {}
        for collator, default_collate_mode in zip(self.collators, self.default_collate_modes):
            if default_collate_mode is None:
                assert not called_default_collate

            if default_collate_mode == "before" and not called_default_collate:
                batch = default_collate(batch)
                called_default_collate = True

            batch = collator.collate(batch, self.dataset_mode, ctx)

            if default_collate_mode == "after":
                assert not called_default_collate
                batch = default_collate(batch)

        if self.return_ctx:
            return batch, ctx
        return batch



class KDCollator:
    @property
    def default_collate_mode(self):
        raise NotImplementedError

    def __call__(self, *_, **__):
        raise RuntimeError("use collators in combination with KDCollator")

    def collate(self, batch, dataset_mode, ctx=None):
        raise NotImplementedError



# adapted from https://github.com/ptrblck/pytorch_misc/blob/master/pytorch_redis.py
import io
import shutil
import subprocess

import redis
import torch

from .cached_dataset import CachedDataset


class RedisDataset(CachedDataset):
    CONNECTION_TRIES = 10
    CUSTOM_ENCODE_TRANSFORMS = []
    CUSTOM_DECODE_TRANSFORMS = []

    @staticmethod
    def _tensor_to_bytes(tensor):
        buffer = io.BytesIO()
        torch.save(tensor, buffer)
        buffer.seek(0)
        return buffer.read()

    @staticmethod
    def _tensor_from_bytes(raw_item):
        buffer = io.BytesIO(raw_item)
        return torch.load(buffer)

    @staticmethod
    def get_encode_transform(sample_item):
        for custom_encode_transform in RedisDataset.CUSTOM_ENCODE_TRANSFORMS:
            data_type_matched, encode_transform = custom_encode_transform(sample_item)
            if data_type_matched:
                return encode_transform
        if isinstance(sample_item, bool):
            return int
        if torch.is_tensor(sample_item):
            return RedisDataset._tensor_to_bytes
        return None

    @staticmethod
    def get_decode_transform(sample_item):
        for custom_decode_transform in RedisDataset.CUSTOM_DECODE_TRANSFORMS:
            data_type_matched, decode_transform = custom_decode_transform(sample_item)
            if data_type_matched:
                return decode_transform
        if isinstance(sample_item, int):
            return int
        if isinstance(sample_item, float):
            return float
        if isinstance(sample_item, str):
            return lambda raw_item: raw_item.decode("utf-8")
        if isinstance(sample_item, bool):
            return lambda b: bool(int(b))
        if torch.is_tensor(sample_item):
            return RedisDataset._tensor_from_bytes
        return None

    def __init__(
            self,
            *args,
            host="localhost",
            port=6379,
            start_if_not_running=True,
            encode_transforms=None,
            decode_transforms=None,
            **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.db = redis.Redis(host=host, port=port)
        self.db_process = None
        for i in range(self.CONNECTION_TRIES):
            try:
                self.db.ping()
                self.db.flushall()
                break
            except redis.exceptions.ConnectionError:
                msg = f"couldn't find redis server on '{host}:{port}'"
                if start_if_not_running:
                    assert shutil.which("redis-server") is not None, "can't find command 'redis-server'"
                    try_start_local_str = " -> trying to start local server"
                    self.logger.info(f"{msg}{try_start_local_str}")
                    try:
                        self.db_process = subprocess.Popen(["redis-server", "--port", f"{port}"])
                        self.logger.info(f"started redis-server on port {port}")
                    except:
                        raise
                else:
                    try_string = "" if i == 0 else f" {i + 1}/{self.CONNECTION_TRIES}"
                    self.logger.info(f"{msg}{try_string}")
                if i == self.CONNECTION_TRIES - 1:
                    raise

        # store how many items are returned from the dataset
        # e.g. ImageNet returns 2 items (image and class label)
        sample = self.dataset[0]
        self.items_per_sample = len(sample)
        # redis encoder doesn't support all datatypes
        self.encode_transforms = encode_transforms
        # redis returns bytes -> apply a decoding transform to every retrieved sample
        self.decode_transforms = decode_transforms

        # initialize encode_transforms
        if self.encode_transforms is None:
            # automatically initialize encode_transforms from known datatypes
            self.encode_transforms = [self.get_encode_transform(item) for item in sample]
        else:
            assert len(self.encode_transforms) == self.items_per_sample
        # initialize decode_transforms
        if self.decode_transforms is None:
            # automatically initialize decode_transforms from known datatypes
            self.decode_transforms = [self.get_decode_transform(item) for item in sample]
        else:
            assert len(self.decode_transforms) == self.items_per_sample

    def _cached_getitem(self, idx):
        db_idx = idx * self.items_per_sample
        if not self.db.exists(db_idx):
            sample = self.dataset[idx]
            for i, (encode_transform, item) in enumerate(zip(self.encode_transforms, sample)):
                encoded_item = encode_transform(item) if encode_transform is not None else item
                self.db.set(db_idx + i, encoded_item)
        else:
            raw_sample = [self.db.get(db_idx + i) for i in range(self.items_per_sample)]
            sample = tuple(
                decode_transform(raw_item) if decode_transform is not None else raw_item
                for decode_transform, raw_item in zip(self.decode_transforms, raw_sample)
            )
        return sample

    def dispose(self):
        self.db.close()
        if self.db_process is not None:
            self.db_process.close()



import logging

from torch.utils.data import Dataset

from kappadata.errors import UseModeWrapperException


class KDDataset(Dataset):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.logger = logging.getLogger(type(self).__name__)

    def __len__(self):
        raise NotImplementedError

    def __enter__(self):
        return self

    def __exit__(self, *_):
        self.dispose()

    def dispose(self):
        """ release resources occupied by dataset (e.g. filehandles) """
        pass

    @property
    def root_dataset(self):
        return self

    @staticmethod
    def has_wrapper(wrapper):
        return False

    @staticmethod
    def has_wrapper_type(wrapper_type):
        return False

    @property
    def all_wrappers(self):
        return []

    @property
    def all_wrapper_types(self):
        return []

    def __getitem__(self, idx):
        raise UseModeWrapperException



import torch

from kappadata.functional import get_random_bbox, cutmix_batch
from .base.mix_collator_base import MixCollatorBase


class MixCollator(MixCollatorBase):
    def __init__(self, cutmix_alpha, mixup_alpha, mode, cutmix_p=.5, **kwargs):
        super().__init__(**kwargs)
        assert isinstance(cutmix_alpha, (int, float)) and 0. < cutmix_alpha
        assert isinstance(mixup_alpha, (int, float)) and 0. < mixup_alpha
        assert isinstance(cutmix_p, (int, float)) and 0. < cutmix_p < 1.
        assert mode in ["sample", "batch"]
        self.cutmix_alpha = cutmix_alpha
        self.mixup_alpha = mixup_alpha
        self.mode = mode
        self.cutmix_p = cutmix_p
        self._is_batch_mode = self.mode == "batch"

    def _collate(self, x, y, batch_size, ctx):
        if self._is_batch_mode:
            use_cutmix_size = 1
        else:
            use_cutmix_size = batch_size
        apply = torch.rand(size=(batch_size,), generator=self.rng) < self.p
        idx2 = torch.from_numpy(self.np_rng.permutation(batch_size)).type(torch.long)
        use_cutmix = torch.rand(size=(use_cutmix_size,), generator=self.rng) < self.cutmix_p

        if ctx is not None:
            ctx["mix_idx2"] = idx2
            ctx["use_cutmix"] = use_cutmix

        if self._is_batch_mode:
            if use_cutmix:
                cutmix_lamb = torch.from_numpy(self.np_rng.beta(self.cutmix_alpha, self.cutmix_alpha, size=batch_size))
                cutmix_lamb = cutmix_lamb.type(torch.float32)
                # TODO batch version
                h, w = x.shape[2:]
                bbox = [get_random_bbox(h=h, w=w, lamb=lamb, rng=self.np_rng) for lamb in cutmix_lamb]
                if ctx is not None:
                    ctx["mix_lambda"] = cutmix_lamb
                    ctx["mix_bbox"] = bbox
                x_clone = x.clone()
                mixed_x = cutmix_batch(x1=x_clone, x2=x_clone[idx2], bbox=bbox)
                mixed_y = self._mix_y(y=y, lamb=cutmix_lamb.view(-1, 1), idx2=idx2)
            else:
                mixup_lamb = torch.from_numpy(self.np_rng.beta(self.mixup_alpha, self.mixup_alpha, size=batch_size))
                mixup_lamb = mixup_lamb.type(torch.float32)
                if ctx is not None:
                    ctx["mix_lambda"] = mixup_lamb
                    ctx["mix_bbox"] = None
                mixup_lamb_x = mixup_lamb.view(-1, *[1] * (x.ndim - 1))
                mixed_x = mixup_lamb_x * x + (1. - mixup_lamb_x) * x[idx2]
                mixed_y = self._mix_y(y=y, lamb=mixup_lamb.view(-1, 1), idx2=idx2)
        else:
            # cutmix params
            cutmix_lamb = torch.from_numpy(self.np_rng.beta(self.cutmix_alpha, self.cutmix_alpha, size=batch_size))
            cutmix_lamb = cutmix_lamb.type(torch.float32)
            # TODO batch version
            h, w = x.shape[2:]
            bbox = [get_random_bbox(h=h, w=w, lamb=lamb, rng=self.np_rng) for lamb in cutmix_lamb]
            # mixup params
            mixup_lamb = torch.from_numpy(self.np_rng.beta(self.mixup_alpha, self.mixup_alpha, size=batch_size))
            mixup_lamb = mixup_lamb.type(torch.float32)

            if ctx is not None:
                ctx["mix_mixup_lambda"] = mixup_lamb
                ctx["mix_cutmix_lambda"] = cutmix_lamb
                ctx["mix_bbox"] = bbox

            mixup_lamb_x = mixup_lamb.view(-1, *[1] * (x.ndim - 1))
            use_cutmix_x = use_cutmix.view(-1, *[1] * (x.ndim - 1))
            x_clone = x.clone()
            cutmixed_x = cutmix_batch(x1=x_clone, x2=x_clone[idx2], bbox=bbox)
            mixuped_x = mixup_lamb_x * x + (1. - mixup_lamb_x) * x[idx2]
            mixed_x = torch.where(use_cutmix_x, cutmixed_x, mixuped_x)

            cutmixed_y = self._mix_y(y=y, lamb=cutmix_lamb.view(-1, 1), idx2=idx2)
            mixuped_y = self._mix_y(y=y, lamb=mixup_lamb.view(-1, 1), idx2=idx2)
            mixed_y = torch.where(use_cutmix.view(-1, 1), cutmixed_y, mixuped_y)

        # filter by apply
        apply_x = apply.view(-1, *[1] * (x.ndim - 1))
        result_x = torch.where(apply_x, mixed_x, x)
        if y is not None:
            apply_y = apply.view(-1, 1)
            result_y = torch.where(apply_y, mixed_y, y)
        else:
            result_y = None

        if result_y is not None:
            return result_x, result_y
        return result_x



import shutil
import zipfile
from collections import namedtuple
from pathlib import Path

CopyResult = namedtuple("CopyFolderResult", "was_copied was_deleted was_zip")


def _log(log, msg):
    if log is not None:
        log(msg)


def copy_folder_from_global_to_local(global_path, local_path, relative_path=None, log=None):
    if not isinstance(global_path, Path):
        global_path = Path(global_path).expanduser()
    if not isinstance(local_path, Path):
        local_path = Path(local_path).expanduser()
    if relative_path is not None and not isinstance(relative_path, Path):
        # relative path can be .zip -> relative path is always without .zip as dst_path is never .zip
        relative_path = Path(relative_path).with_suffix("")

    # check src_path exists (src_path can be folder or .zip)
    src_path = global_path / relative_path if relative_path is not None else global_path
    assert (src_path.exists() and src_path.is_dir()) or src_path.with_suffix(".zip").exists(), \
        f"src_path doesn't exist (can be folder or .zip) '{src_path}'"

    # if dst_path exists:
    # - autocopy start/end file exists -> already copied -> do nothing
    # - autocopy start file exists && autocopy end file doesn't exist -> incomplete copy -> delete and copy again
    # - autocopy start file doesn't exists -> manually copied dataset -> do nothing
    dst_path = local_path / relative_path if relative_path is not None else local_path
    start_copy_file = dst_path / "autocopy_start.txt"
    end_copy_file = dst_path / "autocopy_end.txt"
    was_deleted = False
    if dst_path.exists():
        if start_copy_file.exists():
            if end_copy_file.exists():
                # already automatically copied -> do nothing
                _log(log, f"dataset was already automatically copied '{dst_path}'")
                return CopyResult(was_copied=False, was_deleted=False, was_zip=False)
            else:
                # incomplete copy -> delete and copy again
                _log(log, f"found incomplete automatic copy in '{dst_path}' -> deleting folder")
                shutil.rmtree(dst_path)
                was_deleted = True
                dst_path.mkdir()
        else:
            _log(log, f"using manually copied dataset '{dst_path}'")
            return CopyResult(was_copied=False, was_deleted=False, was_zip=False)
    else:
        dst_path.mkdir(parents=True)

    # create start_copy_file
    with open(start_copy_file, "w") as f:
        f.write("this file indicates that an attempt to copy the dataset automatically was started")

    # copy
    was_zip = False
    if src_path.exists() and src_path.is_dir():
        _log(log, f"copying '{src_path}' to '{dst_path}'")
        # copy folder (dirs_exist_ok=True because dst_path is created for start_copy_file)
        shutil.copytree(src_path, dst_path, dirs_exist_ok=True)
    elif src_path.with_suffix(".zip").exists():
        _log(log, f"extracting '{src_path.with_suffix('.zip')}' to '{dst_path}'")
        # extract zip
        was_zip = True
        with zipfile.ZipFile(src_path.with_suffix(".zip")) as f:
            f.extractall(dst_path)
    else:
        raise NotImplementedError

    # create end_copy_file
    with open(end_copy_file, "w") as f:
        f.write("this file indicates that the copying the dataset automatically was successful")

    _log(log, "finished copying data from global to local")
    return CopyResult(was_copied=True, was_deleted=was_deleted, was_zip=was_zip)



import torch

from kappadata.datasets.kd_dataset import KDDataset


class SequenceIndexDataset(KDDataset):
    def __init__(self, lengths):
        super().__init__()
        self.lengths = lengths

    def getitem_x(self, idx, _=None):
        return torch.arange(idx, idx + self.lengths[idx])

    def __len__(self):
        return len(self.lengths)



import torch
from torchvision.transforms.functional import to_tensor, normalize
from .base.kd_transform import KDTransform

class ImageRangeNorm(KDTransform):
    def __call__(self, x, _=None):
        if not torch.is_tensor(x):
            x = to_tensor(x)
        n_channels = x.size(0)
        values = tuple(0.5 for _ in range(n_channels))
        return normalize(x, mean=values, std=values, inplace=True)



from .base.kd_transform import KDTransform
from .image_pos_embed_grid import ImagePosEmbedGrid
from .image_pos_embed_sincos import ImagePosEmbedSincos
from .image_range_norm import ImageRangeNorm
from .patchify_image import PatchifyImage
from .patchwise_random_rotation import PatchwiseRandomRotation
from .unpatchify_image import UnpatchifyImage



from kappadata.datasets.kd_wrapper import KDWrapper


class XTransformWrapper(KDWrapper):
    def __init__(self, *args, x_transform, **kwargs):
        super().__init__(*args, **kwargs)
        self.x_transform = x_transform

    def getitem_x(self, idx, ctx=None):
        x = self.dataset.getitem_x(idx, ctx)
        transformed = self.x_transform(x)
        return transformed



import numpy as np
import torch

from kappadata.functional import to_onehot_matrix
from .kd_collator import KDCollator


class MixCollatorBase(KDCollator):
    def __init__(self, p=1., n_classes=None, seed=None, **kwargs):
        super().__init__(**kwargs)
        assert isinstance(p, (int, float)) and 0. < p <= 1.
        assert n_classes is None or (isinstance(n_classes, int) and n_classes > 1)
        self.p = p
        self.n_classes = n_classes
        self.rng = torch.Generator()
        if seed is not None:
            self.rng.manual_seed(seed)
        self.np_rng = np.random.default_rng(seed=seed)

    @property
    def default_collate_mode(self):
        return "before"

    def collate(self, batch, dataset_mode, ctx=None):
        if dataset_mode == "x":
            x = batch
            y = None
        elif dataset_mode == "x class":
            x, y = batch
            y = to_onehot_matrix(y, n_classes=self.n_classes).type(torch.float32)
        else:
            raise NotImplementedError
        batch_size = len(x)
        return self._collate(x, y, batch_size, ctx)

    def _collate(self, x, y, batch_size, ctx):
        raise NotImplementedError

    @staticmethod
    def _mix_y(y, lamb, idx2):
        if y is None:
            return y
        lamb_y = lamb.view(-1, 1)
        return lamb_y * y + (1. - lamb_y) * y[idx2]



import unittest

import redis
import torch

from tests_util.redis_mock import RedisMock


class TestRedisMock(unittest.TestCase):
    def test_encode(self):
        kwargs = dict(host="localhost", port=6379)
        og = redis.Redis(**kwargs)
        mock = RedisMock(**kwargs)

        sources = [5, 10., "test", b"102", torch.tensor([1., 2.])]
        for i, source in enumerate(sources):
            og.set(i, source)
            mock.set(i, source)
            self.assertTrue(og.exists(i))
            self.assertTrue(mock.exists(i))
            og_actual = og.get(i)
            mock_actual = og.get(i)
            self.assertEqual(og_actual, mock_actual)



import numpy as np
from torch.nn.functional import one_hot


def get_random_bbox(h, w, lamb, rng):
    # TODO batch version of this
    #  (should use torch for it but torch.randint might have different impl than rng.intergers)
    cut_ratio = np.sqrt(1. - lamb)
    cut_h = int(h * cut_ratio)
    cut_w = int(w * cut_ratio)

    h_center = rng.integers(h)
    w_center = rng.integers(w)

    left = np.clip(w_center - cut_w // 2, 0, w)
    right = np.clip(w_center + cut_w // 2, 0, w)
    top = np.clip(h_center - cut_h // 2, 0, h)
    bot = np.clip(h_center + cut_h // 2, 0, h)

    return top, left, bot, right


def get_area_of_bbox(bbox, h, w):
    top, left, bot, right = bbox
    return 1 - (right - left) * (bot - top) / (h * w)


def to_onehot_vector(y, n_classes):
    if y.ndim == 0:
        y = one_hot(y, num_classes=n_classes)
    return y


def to_onehot_matrix(y, n_classes):
    if y.ndim == 1:
        y = one_hot(y, num_classes=n_classes)
    return y


def cutmix_batch(x1, x2, bbox):
    for i in range(len(x1)):
        cutmix_single(x1=x1[i], x2=x2[i], bbox=bbox[i])
    return x1


def cutmix_single(x1, x2, bbox):
    top, left, bot, right = bbox
    x1[:, top:bot, left:right] = x2[:, top:bot, left:right]
    return x1



import numpy as np

from kappadata.datasets.kd_subset import KDSubset


class SubsetWrapper(KDSubset):
    def __init__(self, dataset, indices=None, start_index=None, end_index=None):
        if indices is None:
            # create indices from start/end index
            assert start_index is not None or end_index is not None
            assert start_index is None or isinstance(start_index, int)
            assert end_index is None or isinstance(end_index, int)
            end_index = end_index or len(dataset)
            end_index = min(end_index, len(dataset))
            start_index = start_index or 0
            indices = np.arange(start_index, end_index)
        else:
            assert start_index is None and end_index is None
        super().__init__(dataset=dataset, indices=indices)



from .base.mix_wrapper_base import MixWrapperBase


class MixupWrapper(MixWrapperBase):
    def __init__(self, *args, alpha, **kwargs):
        super().__init__(*args, **kwargs)
        assert isinstance(alpha, (int, float)) and 0. < alpha
        self.alpha = alpha

    def _get_params(self, ctx):
        if ctx is None or "mixup_apply" not in ctx:
            apply = self.rng.random() < self.p
            if ctx is not None:
                ctx["mixup_apply"] = apply
            if apply:
                lamb = self.rng.beta(self.alpha, self.alpha)
                idx2 = self.rng.integers(0, len(self))
                if ctx is not None:
                    ctx["mixup_lambda"] = lamb
                    ctx["mixup_idx2"] = idx2
            else:
                ctx["mixup_lambda"] = -1
                ctx["mixup_idx2"] = -1
                return False, None, None
        else:
            apply = ctx["mixup_apply"]
            if apply:
                lamb = ctx["mixup_lambda"]
                idx2 = ctx["mixup_idx2"]
            else:
                return False, None, None
        return True, lamb, idx2

    def getitem_x(self, idx, ctx=None):
        apply, lamb, idx2 = self._get_params(ctx)
        x1 = self.dataset.getitem_x(idx, ctx)
        if not apply:
            return x1
        x2 = self.dataset.getitem_x(idx2, ctx)
        return lamb * x1 + (1. - lamb) * x2

    def getitem_class(self, idx, ctx=None):
        apply, lamb, idx2 = self._get_params(ctx)
        y1 = self._getitem_class(idx, ctx)
        if not apply:
            return y1
        y2 = self._getitem_class(idx2, ctx)
        return lamb * y1 + (1. - lamb) * y2



import numpy as np
import torch

from kappadata.datasets.kd_wrapper import KDWrapper
from kappadata.functional import to_onehot_vector


class MixWrapperBase(KDWrapper):
    def __init__(self, *args, p=1., seed=None, **kwargs):
        super().__init__(*args, **kwargs)
        assert isinstance(p, (int, float)) and 0. < p <= 1.
        self.p = p
        self.rng = np.random.default_rng(seed=seed)

    def _getitem_class(self, idx, ctx):
        y = self.dataset.getitem_class(idx, ctx)
        if not torch.is_tensor(y):
            y = torch.tensor(y)
        return to_onehot_vector(y, n_classes=self.dataset.n_classes)



import torch

from .base.mix_collator_base import MixCollatorBase


class MixupCollator(MixCollatorBase):
    def __init__(self, alpha, **kwargs):
        super().__init__(**kwargs)
        assert isinstance(alpha, (int, float)) and 0. < alpha
        self.alpha = alpha

    def _collate(self, x, y, batch_size, ctx):
        apply = torch.rand(size=(batch_size,), generator=self.rng) < self.p
        lamb = torch.from_numpy(self.np_rng.beta(self.alpha, self.alpha, size=batch_size)).type(torch.float32)
        idx2 = torch.from_numpy(self.np_rng.permutation(batch_size)).type(torch.long)

        if ctx is not None:
            ctx["mixup_lambda"] = lamb
            ctx["mixup_idx2"] = idx2

        # add dimensions for broadcasting
        apply_x = apply.view(-1, *[1] * (x.ndim - 1))
        lamb_x = lamb.view(-1, *[1] * (x.ndim - 1))

        mixed_x = lamb_x * x + (1. - lamb_x) * x[idx2]
        result_x = torch.where(apply_x, mixed_x, x)
        mixed_y = self._mix_y(y=y, lamb=lamb, idx2=idx2)
        result_y = torch.where(apply.view(-1, 1), mixed_y, y)
        if result_y is not None:
            return result_x, result_y
        return result_x



import unittest

import torch

from kappadata.transforms.patchify_image import PatchifyImage
from kappadata.transforms.unpatchify_image import UnpatchifyImage


class TestPatchifyUnpatchifyImage(unittest.TestCase):
    def test_squareimg_squarepatch(self):
        x = torch.randn(4, 8, 8)
        forward = PatchifyImage(patch_size=2)
        backward = UnpatchifyImage()
        ctx = {}
        self.assertTrue(torch.all(x == backward(forward(x, ctx), ctx)))

    def test_squareimg_rectpatch(self):
        x = torch.randn(4, 8, 8)
        forward = PatchifyImage(patch_size=(4, 2))
        backward = UnpatchifyImage()
        ctx = {}
        self.assertTrue(torch.all(x == backward(forward(x, ctx), ctx)))

    def test_rectimg_squarepatch(self):
        x = torch.randn(4, 8, 16)
        forward = PatchifyImage(patch_size=2)
        backward = UnpatchifyImage()
        ctx = {}
        self.assertTrue(torch.all(x == backward(forward(x, ctx), ctx)))

    def test_rectimg_rectpatch(self):
        x = torch.randn(4, 8, 16)
        forward = PatchifyImage(patch_size=(4, 2))
        backward = UnpatchifyImage()
        ctx = {}
        self.assertTrue(torch.all(x == backward(forward(x, ctx), ctx)))



import torch
from torchvision.transforms.functional import rotate

from .base.kd_transform import KDTransform


class PatchwiseRandomRotation(KDTransform):
    def __init__(self, seed=None):
        self.rng = torch.Generator()
        if seed is not None:
            self.rng.manual_seed(seed)

    def __call__(self, x, ctx=None):
        c, l, patch_h, patch_w = x.shape
        # sample rotations
        rotations = torch.randint(0, 4, size=(l,), generator=self.rng) * 90
        if ctx is not None:
            ctx["patchwise_random_rotation"] = rotations
        # rotate patches
        for i in range(l):
            x[:, i] = rotate(x[:, i], angle=rotations[i].item())
        return x



from kappadata.datasets.kd_dataset import KDDataset


class ClassificationDataset(KDDataset):
    def __init__(self, x, classes):
        super().__init__()
        assert len(x) == len(classes)
        self.x = x
        self.classes = classes

    def getitem_x(self, idx, _=None):
        return self.x[idx].clone()

    def getitem_class(self, idx, _=None):
        return self.classes[idx]

    @property
    def n_classes(self):
        return max(self.classes) + 1

    def __len__(self):
        return len(self.classes)



import unittest

from kappadata.wrappers.dataset_wrappers.percent_filter_wrapper import PercentFilterWrapper
from kappadata.wrappers.dataset_wrappers.repeat_wrapper import RepeatWrapper
from kappadata.wrappers.dataset_wrappers.shuffle_wrapper import ShuffleWrapper
from kappadata.wrappers.sample_wrappers.mixup_wrapper import MixupWrapper
from tests_util.index_dataset import IndexDataset


class TestAllWrapperTypes(unittest.TestCase):
    def test(self):
        root_ds = IndexDataset(size=10)
        wrapper1 = PercentFilterWrapper(root_ds, from_percent=0.3)
        wrapper2 = RepeatWrapper(wrapper1, repetitions=2)
        wrapper3 = MixupWrapper(wrapper2, alpha=1., p=1.)
        wrapper4 = ShuffleWrapper(wrapper3, seed=2)
        wrapper5 = PercentFilterWrapper(wrapper4, to_percent=0.5)

        self.assertEqual(
            [PercentFilterWrapper, ShuffleWrapper, MixupWrapper, RepeatWrapper, PercentFilterWrapper],
            wrapper5.all_wrapper_types,
        )
        self.assertEqual(
            [ShuffleWrapper, MixupWrapper, RepeatWrapper, PercentFilterWrapper],
            wrapper4.all_wrapper_types,
        )
        self.assertEqual(
            [ShuffleWrapper, MixupWrapper, RepeatWrapper, PercentFilterWrapper],
            wrapper4.all_wrapper_types,
        )
        self.assertEqual([MixupWrapper, RepeatWrapper, PercentFilterWrapper], wrapper3.all_wrapper_types)
        self.assertEqual([RepeatWrapper, PercentFilterWrapper], wrapper2.all_wrapper_types)
        self.assertEqual([PercentFilterWrapper], wrapper1.all_wrapper_types)



import torch

from kappadata.functional import get_random_bbox, get_area_of_bbox, cutmix_single
from .base.mix_wrapper_base import MixWrapperBase


class CutmixWrapper(MixWrapperBase):
    def __init__(self, *args, alpha, **kwargs):
        super().__init__(*args, **kwargs)
        assert isinstance(alpha, (int, float)) and 0. < alpha
        self.alpha = alpha

    @staticmethod
    def _context_has_params(ctx):
        return ctx is not None and "cutmix_apply" in ctx

    def _get_params(self, ctx, h=None, w=None):
        if not self._context_has_params(ctx):
            apply = self.rng.random() < self.p
            if ctx is not None:
                ctx["cutmix_apply"] = apply
            if apply:
                lamb = self.rng.beta(self.alpha, self.alpha)
                idx2 = self.rng.integers(0, len(self))
                bbox = get_random_bbox(h=h, w=w, lamb=lamb, rng=self.rng)
                # correct lambda to actual area
                lamb = get_area_of_bbox(bbox=bbox, h=h, w=w)
                if ctx is not None:
                    ctx["cutmix_lambda"] = lamb
                    ctx["cutmix_idx2"] = idx2
                    ctx["cutmix_bbox"] = bbox
            else:
                ctx["cutmix_lambda"] = -1
                ctx["cutmix_idx2"] = -1
                ctx["cutmix_bbox"] = (-1, -1, -1, -1)
                return False, None, None, None
        else:
            apply = ctx["cutmix_apply"]
            if apply:
                lamb = ctx["cutmix_lambda"]
                idx2 = ctx["cutmix_idx2"]
                bbox = ctx["cutmix_bbox"]
            else:
                return False, None, None, None
        return True, lamb, idx2, bbox

    def getitem_x(self, idx, ctx=None):
        x1 = self.dataset.getitem_x(idx, ctx)
        assert torch.is_tensor(x1) and x1.ndim == 3, "convert image to tensor before CutmixWrapper"
        h, w = x1.shape[1:]
        apply, lamb, idx2, bbox = self._get_params(ctx=ctx, h=h, w=w)
        if not apply:
            return x1
        x2 = self.dataset.getitem_x(idx2, ctx)
        assert x1.shape == x2.shape
        return cutmix_single(x1=x1, x2=x2, bbox=bbox)

    def getitem_class(self, idx, ctx=None):
        if self._context_has_params(ctx):
            apply, lamb, idx2, _ = self._get_params(ctx=ctx)
        else:
            h, w = self.getitem_x(idx, ctx).shape[1:]
            apply, lamb, idx2, _ = self._get_params(ctx=ctx, h=h, w=w)
        y1 = self._getitem_class(idx, ctx)
        if not apply:
            return y1
        y2 = self._getitem_class(idx2, ctx)
        return lamb * y1 + (1. - lamb) * y2



from redis.connection import Encoder


class RedisMock:
    def __init__(self, host, port):
        self.host = host
        self.port = port
        self.is_closed = False
        self.items = {}
        self.encoder = Encoder(
            encoding="utf-8",
            encoding_errors="strict",
            decode_responses=False,
        )

    def exists(self, key):
        return key in self.items

    def set(self, key, value):
        encoded = self.encoder.encode(value)
        self.items[key] = encoded

    def get(self, key):
        return self.items[key]

    def close(self):
        self.is_closed = True

    def get_encoder(self):
        return self.encoder

    def ping(self):
        pass

    def flushall(self):
        self.items.clear()



import unittest

import torch

from kappadata.wrappers.sample_wrappers.cutmix_wrapper import CutmixWrapper
from kappadata.wrappers.sample_wrappers.mixup_wrapper import MixupWrapper
from tests_util.classification_dataset import ClassificationDataset


class TestCutmixMixupCombination(unittest.TestCase):
    def test_getitem_class_manual(self):
        rng = torch.Generator().manual_seed(42)
        data = torch.randn(4, 1, 16, 16, generator=rng)
        classes = torch.randint(4, size=(4,), generator=rng)
        ds = ClassificationDataset(x=data, classes=classes)
        mix_ds = MixupWrapper(dataset=CutmixWrapper(dataset=ds, alpha=1., p=1., seed=101), alpha=1., p=1., seed=1304)

        y0 = mix_ds.getitem_class(0)
        y1 = mix_ds.getitem_class(1)
        self.assertTrue(torch.allclose(torch.tensor([0., 0.9577113986, 0.0006607597, 0.0416278653]), y0))
        self.assertTrue(torch.allclose(torch.tensor([0., 0.9795139432, 0.0204860158, 0.]), y1))

    def test_getitem_class_automatic_noctx(self):
        rng = torch.Generator().manual_seed(42)
        data = torch.randn(16, 1, 8, 8, generator=rng)
        classes = torch.randint(4, size=(len(data),), generator=rng)
        ds = ClassificationDataset(x=data, classes=classes)
        mix_ds = MixupWrapper(dataset=CutmixWrapper(dataset=ds, alpha=1., p=1., seed=101), alpha=1., p=1., seed=1304)

        max_nonzero_class_prob_count = 0
        for i in range(len(data)):
            y = mix_ds.getitem_class(i)
            self.assertTrue(torch.allclose(torch.tensor(1.), y.sum()))
            nonzero_class_prob_count = (y != 0.).sum()
            max_nonzero_class_prob_count = max(max_nonzero_class_prob_count, nonzero_class_prob_count)
            self.assertLessEqual(nonzero_class_prob_count, 4)
        self.assertEqual(4, max_nonzero_class_prob_count)



import torch

from kappadata.functional import get_random_bbox, get_area_of_bbox, cutmix_single
from .base.mix_wrapper_base import MixWrapperBase


class MixWrapper(MixWrapperBase):
    def __init__(self, *args, cutmix_alpha, mixup_alpha, cutmix_p=.5, **kwargs):
        super().__init__(*args, **kwargs)
        assert isinstance(cutmix_alpha, (int, float)) and 0. < cutmix_alpha
        assert isinstance(mixup_alpha, (int, float)) and 0. < mixup_alpha
        assert isinstance(cutmix_p, (int, float)) and 0. < cutmix_p < 1.
        self.cutmix_alpha = cutmix_alpha
        self.mixup_alpha = mixup_alpha
        self.cutmix_p = cutmix_p

    def _get_apply_and_usecutmix(self, ctx):
        if ctx is None or "mix_apply" not in ctx:
            apply = self.rng.random() < self.p
            if apply:
                use_cutmix = self.rng.random() < self.cutmix_p
            else:
                use_cutmix = False
            if ctx is not None:
                ctx["mix_apply"] = apply
                ctx["mix_usecutmix"] = use_cutmix
                if not apply:
                    ctx["mix_idx2"] = -1
                    ctx["mix_lambda"] = -1
                    ctx["mix_bbox"] = (-1, -1, -1, -1)
        else:
            apply = ctx["mix_apply"]
            use_cutmix = ctx["mix_usecutmix"]
        return apply, use_cutmix

    def _get_idx2(self, ctx):
        if ctx is not None and "mix_idx2" in ctx:
            return ctx["mix_idx2"]
        idx2 = self.rng.integers(0, len(self))
        if ctx is not None:
            ctx["mix_idx2"] = idx2
        return idx2

    def _get_cutmix_params(self, ctx, h=None, w=None):
        if ctx is None or "mix_lambda" not in ctx:
            lamb = self.rng.beta(self.cutmix_alpha, self.cutmix_alpha)
            bbox = get_random_bbox(h=h, w=w, lamb=lamb, rng=self.rng)
            # correct lambda to actual area
            lamb = get_area_of_bbox(bbox=bbox, h=h, w=w)
            if ctx is not None:
                ctx["mix_lambda"] = lamb
                ctx["mix_bbox"] = bbox
        else:
            lamb = ctx["mix_lambda"]
            bbox = ctx["mix_bbox"]
        return lamb, bbox

    def _get_mixup_params(self, ctx):
        if ctx is None or "mix_lambda" not in ctx:
            lamb = self.rng.beta(self.cutmix_alpha, self.cutmix_alpha)
            if ctx is not None:
                ctx["mix_lambda"] = lamb
                ctx["mix_bbox"] = (-1, -1, -1, -1)
        else:
            lamb = ctx["mix_lambda"]
        return lamb

    def getitem_x(self, idx, ctx=None):
        apply, use_cutmix = self._get_apply_and_usecutmix(ctx)
        x1 = self.dataset.getitem_x(idx, ctx)
        if not apply:
            return x1
        idx2 = self._get_idx2(ctx)
        x2 = self.dataset.getitem_x(idx2, ctx)
        if use_cutmix:
            assert torch.is_tensor(x1) and x1.ndim == 3, "convert image to tensor before MixWrapper"
            assert x1.shape == x2.shape
            h, w = x1.shape[1:]
            lamb, bbox = self._get_cutmix_params(ctx=ctx, h=h, w=w)
            return cutmix_single(x1=x1, x2=x2, bbox=bbox)
        else:
            lamb = self._get_mixup_params(ctx=ctx)
            return lamb * x1 + (1. - lamb) * x2

    def getitem_class(self, idx, ctx=None):
        apply, use_cutmix = self._get_apply_and_usecutmix(ctx)
        if not apply:
            return self._getitem_class(idx, ctx)
        if use_cutmix:
            if ctx is not None and "mix_bbox" in ctx:
                lamb, _ = self._get_cutmix_params(ctx=ctx)
            else:
                h, w = self.getitem_x(idx, ctx).shape[1:]
                lamb, _ = self._get_cutmix_params(ctx=ctx, h=h, w=w)
        else:
            lamb = self._get_mixup_params(ctx)
        y1 = self._getitem_class(idx, ctx)
        idx2 = self._get_idx2(ctx)
        y2 = self._getitem_class(idx2, ctx)
        return lamb * y1 + (1. - lamb) * y2



from kappadata.datasets.kd_dataset import KDDataset


class ClassDataset(KDDataset):
    def __init__(self, classes):
        super().__init__()
        self.classes = classes

    def getitem_class(self, idx, _=None):
        return self.classes[idx]

    @property
    def n_classes(self):
        return max(self.classes) + 1

    def __len__(self):
        return len(self.classes)



import torch

from kappadata.datasets.kd_dataset import KDDataset


class SequenceClassificationDataset(KDDataset):
    def __init__(self, classes):
        super().__init__()
        self.classes = classes
        self.lengths = [len(c) for c in classes]

    def getitem_x(self, idx, _=None):
        return torch.arange(idx, idx + self.lengths[idx])

    def getitem_classes(self, idx, _=None):
        return self.classes[idx]

    def getitem_seqlen(self, idx, _=None):
        return self.lengths[idx]

    def __len__(self):
        return len(self.lengths)



import unittest

import torch

from kappadata.transforms.patchify_image import PatchifyImage
from kappadata.transforms.patchwise_random_rotation import PatchwiseRandomRotation
from kappadata.transforms.unpatchify_image import UnpatchifyImage


class TestPatchifyRotateUnpatchifyImage(unittest.TestCase):
    def test(self):
        #  1  2  3  4
        #  5  6  7  8
        #  9 10 11 12
        # 13 14 15 16
        source = torch.arange(16).view(1, 4, 4) + 1
        # seed=6 -> rotations=[2, 1, 3, 0]
        # top_left:  [ 1,  2] -> [ 6,  5]
        #            [ 5,  6] -> [ 2,  1]
        # top_right: [ 3,  4] -> [ 4,  8]
        #            [ 7,  8] -> [ 3,  7]
        # bot_left:  [ 9, 10] -> [13,  9]
        #            [13, 14] -> [14, 10]
        # bot_right: [11, 12]
        #            [15, 16]
        target = torch.tensor([6, 5, 4, 8, 2, 1, 3, 7, 13, 9, 11, 12, 14, 10, 15, 16]).view(1, 4, 4)
        forward = PatchifyImage(patch_size=2)
        rotate = PatchwiseRandomRotation(seed=6)
        backward = UnpatchifyImage()
        ctx = {}
        self.assertTrue(torch.all(target == backward(rotate(forward(source, ctx), ctx), ctx)))
